											(* recolleur de gnome *)

(* 
   Les brins d'ADN se rencontrent sous forme d'une pelotte, un fil de ~2 mtres de long y est
   enroul. Pour les tudier il faut commencer par les drouler, ce faisant... il y a des accidents
   de manipulation et on les casse. Ci-dessous un morceau de longueur N est cass en K... si on n'a
   que cela on a (pour k=4) 24*16 = 384 possibilits de brins initiaux

	Ci-dessous on suppose que le mme brin a t cass deux fois, on est donc sr de pouvoir
	trouver au moins deux r-arrangements possibles. On veut les trouver tous.

	Un algorithme joyeux consiste  fabriquer les M=384 brins possibles pour chacun de nos
	K-uplets. Puis on fait les 384 = 147456 comparaisons... le fait que leur longueur soit N
	n'est mme plus important !

	En tant un peu plus malin, au lieu de tous les comparer, on peut les trier, puis comparer par
	exemple par fusion : le cot sera ~ 2 M ln(M) vers 6600
*)
																
type ADN=A|T|C|G;;
let N=100;;		(* la longueur du morceau complet *)
let K=4;;		(* le nombre de morceaux aprs cassures *)

#open "random";; #open "sort";;

(* des utilitaires pour les tirages au sort *)
(* La fonction tirage_a_permutation va rpartir alatoirement les k entiers 0..(k-1) dans un
	vecteur. L'ide est de travailler avec de string : on place 1 seul "au centre". Puis on insre 2
	soit  droite, soit  gauche selon le rsultat d'un int 1, puis on insre 3 en position 0, 1 ou 2
	selon le rsultat d'un int 3....  Une permutation est ainsi code par un tableau [|m(2);...m(n)|]

   Note1 : il s'agit de l'criture des entiers en "base factorielle"
   Note2 : si on travaille avec des entiers de 0..9, la technique d'insertion par chane de caractres
   s'impose, hlas caml n'a pas de "insert" tout prt pour les string *)

let installe nb liste place=
  let rec met list k=
	 if k=0 
	 then nb::list
	 else (hd list):: (met (tl list) (k-1)) 
  in met liste place;;	

let tirage()=
	let liste=make_vect (K-1) 0 in
	for k=0 to (K-2) do liste.(k)<-int (k+2) done;
	liste;;

let tirage_a_permutation()=
  let prepare=tirage() and evolue=ref [1] in
	 for k=2 to K do evolue:=(installe k !evolue prepare.(k-2)) done;
	 vect_of_list !evolue;;

let dedouble x = let lx=(list_of_vect x) in (lx, rev lx);;
let rec flatten = function 
  |[]			->	[]
  |(a,b)::q	->	a::b::(flatten q);;
	
		(*  *)

let dieu () = let x = (int 4) in match x with	|0 -> A |1-> T |2-> C |3->G;;
				
let test = let aa=make_vect 100 A in for k=0 to 99 do aa.(k)<-dieu() done; aa;;
			
let cassures () = let bon = ref false and a=ref 0 and b=ref 0 and c=ref 0 in 
  while not !bon do 
	 a:=5+(int 90);b:=5+(int 90);c:=5+(int 90);
	 bon := (abs(!b - !a)>=5) & (abs(!b - !c)>=5) & (abs(!c - !a)>=5)
  done;
  vect_of_list (sort (prefix <) [!a;!b;!c]);;		

let morceaux () = 	let c=cassures ()
					      and sens=[|int 2;int 2;int 2;int 2|] 
in 	let a=sub_vect test 0 c.(0) 
		and b=sub_vect test c.(0) (c.(1)-c.(0))			
		and c=sub_vect test c.(1) (c.(2)-c.(1))
		and d=sub_vect test c.(2) ((N-1)-c.(2))
in [if sens.(0)=0 then a else (vect_of_list (rev (list_of_vect a)));
	 if sens.(1)=0 then b else (vect_of_list(rev (list_of_vect b)));	
	 if sens.(2)=0 then c else (vect_of_list (rev (list_of_vect c)));
	 if sens.(3)=0 then d else (vect_of_list (rev (list_of_vect d)))];;

(* pour fabriquer un exercice pour se mettre en route :			*)		
let p=morceaux();;
let s=morceaux();;

(* Test " la main" *)
(* On est parti de deux exemplaires d'un mme vecteur de longueur 100 form de A,C,G,T chacun des
	deux a t coup en 4 morceaux de longueurs au moins 5 les morceaux se sont peut-tre retourns
	
	Il faut trouver un (tous!?!) les moyens de reconstituer une liste de longueur 100 qui ait pu
	donner naissance  ces deux quadruplets *)
	
(* premier quadruplet : P, second : S *)

let p_vrac= let vrac=tirage_a_permutation() and vp = (vect_of_list p)
in map_vect (function i->vp.(vrac.(i-1)-1)) [|1;2;3;4|]  ;;
					
let s_vrac= 	let vrac=tirage_a_permutation() and vp = (vect_of_list s)
in map_vect (function i->vp.(vrac.(i-1)-1)) [|1;2;3;4|] ;;
     
(* pour les manipulations  la main : p_vrac_liste et q_vrac_liste  *)       
let pvl = map_vect list_of_vect p_vrac;;
let svl = map_vect list_of_vect s_vrac;;

(* nombre a priori de possibilits : 24*16 pour chaque  soit 147456 	*)  
(* parmi les 8 extrmits de P et les 8 de S se trouvent ncssairement les extrmits du vecteur de
   dpart.  S'il n'y a pas de coincidences il faudra associer ces deux couples. Partant de l, s'il n'y
   a pas de coincidence, on aura un reliquat de longueur qui sera "uniquement" un autre dbut, et on
   continue ainsi S'il y a coincidence, on fait pareil pour chacun des cas *)

let p_debut=flatten (list_of_vect (map_vect dedouble p_vrac));;
let s_debut=flatten (list_of_vect  (map_vect dedouble s_vrac));;

(* on commence par crire une fonction qui prend deux listes et dit si l'une est le dbut de
	l'autre.  Dans le type rponse on peut avoir E(gal, P(remier, S(econd, Non La liste qui suit
	est le bout qui dpasse *)

type reponse = E|P|S|Non;;

let rec inclus = fun
  |[] [] -> E, []
  |[] liste	->	P, liste
  |liste []	->	S, liste
  |(t1::q1) (t2::q2)	->	
     if t1=t2
	  then inclus q1 q2 
	  else  Non,[];;		

(*		inclus  [5;6] [5;6;7];;
inclus  [5;6;7;8;9;10] [5;6;7];;				
inclus  [5;6;7;8;9] [5;6;7;8;9];;				
inclus  [5;6;7;8;9] [4;5;6;7;8;9];;
- : reponse * int list = P, [7]
#- : reponse * int list = S, [8; 9; 10]
#- : reponse * int list = E, []
#- : reponse * int list = Non, []		*)


(* 	bimap (fun a b -> 10*a+b) [1;2;3] [4;8];;	- : int list list = [[14; 18]; [24; 28]; [34; 38]] *)
let bimap f l1 l2=map (function x-> map (f x) l2) l1;;

(* colle [[4;5];[6];[];[9;7;1]];;     - : int list = [4; 5; 6; 9; 7; 1]		*)
let rec colle =  function  |[] -> []  |t::q -> t@(colle q);;	

(* on tente toute les associations d'un p et d'un q *)
colle (bimap inclus p_debut s_debut);;

(* comme l'a montr l'exemple prcdent, le rsultat renseigne bien peu! on r-crit donc une fonction "contenu" qui rpond en informant *)
let contenu  l1 l2 = let (a,b)=(inclus l1 l2) 
in match a with
  |P	->	[a,b,l1,l2] 
  |S -> 	[a,b,l1,l2] 
  |_ -> [];;
		
colle (bimap contenu p_debut s_debut);;
		
(* retire servira quand on aura match des bouts de liste,  enlever lesdits bouts et leurs symtriques, attention ! il ne faut l'enlever qu'une fois *)
	
let retire x liste= 
  let rx = rev x in
  let rec retir = function
	 |[]		->	[] 	
	 |t::q	->	if t=x or t=rx 
		then retir q
		else t::(retir q)
  in retir liste ;;											

(* contenu2 fait comme contenu, mais en mme temps les strings trouvs sont enlevs *)
	
let contenu2 l1 l2 liste_p liste_s= let (a,b)=(inclus l1 l2) in match a with
  |P	->	[a,b,l1,l2,retire l1 liste_p,retire l2 liste_s] 
  |S -> [a,b,l1,l2,retire l1 liste_p,retire l2 liste_s] 
  |_ -> [];;					

colle (bimap (fun a b -> contenu2 a b p_debut s_debut) p_debut s_debut);;

(* vire_vide enlve les listes vides d'une liste de listes *)
let rec vire_vide = function
  |[]		->		[]
  |t::q	->		
     if t=[] 
	  then vire_vide q
	  else t::(vire_vide q);;	

(* A un instant typique, "gnrique", on aura une liste de
   (Lettre,Longueur,debut_p,debut_s,reste_p,reste_s)

				o debut_p et debut_s ont un dbut commun
				si Lettre=P : c'est dbut_p qui est le plus court et dbut_s dpasse de Longueur							
				si Lettre=S : c'est dbut_s qui est le plus court et dbut_s dpasse de Longueur							
				si Lettre=E : dbut_p = dbut_s 
				reste_p et reste_s : les morceaux qui restent (chacun dans les deux sens)
	
	l'action pour continuer sera selon Lettre 
   P : on va aller chercher dans reste_p la partie de dbut_s qui dpasse (ou au moins un dbut) et on
	met  jour
	S : symtrique
	E : pour chacun des redmarrages possibles on met dans la liste un nouveau quintuplet	
	
	On gagnerait de l'espace en codant les essais [[2,5],[3,6,1]] pour dsigner les associations de
	p2,p5 qui commence comme s3,s6,s1 et comme il n'y a finalement pas trop de cas  tester on s'en
	sortirait en essayant tout!  *)

(*
CE QUI SUIT EST NON-FINI 
let continue 	(lettre,longueur,debut_p,debut_s,reste_p,reste_s) = match lettre  with
	|P	->	 let a_trouver = "ce qui depasse" in 
	|S
	|E
	|Non		->	failwith "on n'aurait pas d tre ici"
	
*)	
		
		

(*  l'exemple "manuel"  *)
(*  p_vrac : ADN vect vect = 

 [|[|C; C; T; G; T; G; C; T; A; A; A; T; C; C; C; C; T; T; C; G; A; A; T; G;
     T; T; A|];
   [|C; C; C; T; T; C; G; C; A; A; C; C; T; C; T; T; T; G; G; A; G; A; G; T;
     G; G; A; G; T; A; C; A; T; C; A; G; G; T; G; A; A; G|];
   [|A; A; A; T; G|];
   [|C; T; C; A; G; T; T; A; A; A; A; T; C; T; T; G; A; C; G; T; G; A; A; C;
     A|]|]

 s_vrac : ADN vect vect = 

 [|[|G; T; A; A; A; A; C; A; A; G; T; G; C; A; G; T; T; C; T; A; A; A; A; T;
     T; G; A; C; T; C; G; A; A; G|];
   [|G; C; A; A; C; C; T|];
   [|C; C; T; G; T; G; C; T; A; A; A; T; C; C; C; C; T; T; C; G; A; A; T; G;
     T; T; A; C; C; C; T; T; C|];
   [|C; T; T; T; G; G; A; G; A; G; T; G; G; A; G; T; A; C; A; T; C; A; G; G;
     T|]|]

(* le rsultat sur cet exemple 

(rev pvl.(2))@(rev pvl.(3))@(rev pvl.(1))@(rev pvl.(0));;
(qvl.(0))@(rev qvl.(3))@(rev qvl.(1))@(rev qvl.(2));;
- : ADN list =
 [G; T; A; A; A; A; C; A; A; G; T; G; C; A; G; T; T; C; T; A; A; A; A; T; T;
  G; A; C; T; C; G; A; A; G; T; G; G; A; C; T; A; C; A; T; G; A; G; G; T; G;
  A; G; A; G; G; T; T; T; C; T; C; C; A; A; C; G; C; T; T; C; C; C; A; T; T;
  G; T; A; A; G; C; T; T; C; C; C; C; T; A; A; A; T; C; G; T; G; T; C; C]
#- : ADN list =
 [G; T; A; A; A; A; C; A; A; G; T; G; C; A; G; T; T; C; T; A; A; A; A; T; T;
  G; A; C; T; C; G; A; A; G; T; G; G; A; C; T; A; C; A; T; G; A; G; G; T; G;
  A; G; A; G; G; T; T; T; C; T; C; C; A; A; C; G; C; T; T; C; C; C; A; T; T;
  G; T; A; A; G; C; T; T; C; C; C; C; T; A; A; A; T; C; G; T; G; T; C; C]		
  
 c'est bien sr plus simple dans l'autre sens, mais je suis parti par le "mauvais bout"   	 *)
  



